import itertools
import warnings
from collections import defaultdict, namedtuple
from pathlib import Path

import numpy as np

import pysubgroup as ps


def identity(x, *args, **kwargs):  # pylint: disable=unused-argument
    return x


class GpGrowth:
    def __init__(self, mode="b_u"):
        self.GP_node = namedtuple(
            "GP_node", ["cls", "id", "parent", "children", "stats"]
        )
        self.minSupp = 10
        self.tqdm = identity
        self.depth = 0
        self.mode = mode  # specify eihther b_u (bottom up) or t_d (top down)
        self.constraints_monotone = []
        self.results = []
        self.task = []
        # Future: There also is the option of a stable mode
        # which never creates the prefix trees

    def prepare_selectors(self, search_space, data):
        selectors = []
        assert len(search_space) > 0, "Provided searchspace was empty"
        for selector in search_space:
            cov_arr = selector.covers(data)
            selectors.append((np.count_nonzero(cov_arr), selector, cov_arr))

        selectors = [
            (size, selector, arr)
            for size, selector, arr in selectors
            if all(
                constraint.is_satisfied(arr, slice(None), data)
                for constraint in self.constraints_monotone
            )
        ]
        sorted_selectors = sorted(selectors, reverse=True)
        self.remove_selectors_with_low_optimistic_estimate(
            sorted_selectors, len(search_space)
        )

        selectors_sorted = [selector for size, selector, arr in sorted_selectors]
        if len(selectors_sorted) == 0:
            arrs = np.empty((0, 0), dtype=np.bool_)
        else:
            arrs = np.vstack([arr for size, selector, arr in sorted_selectors]).T
        # print(selectors_sorted)
        return selectors_sorted, arrs

    def remove_selectors_with_low_optimistic_estimate(self, s, search_space_size):
        if not hasattr(self.task.qf, "optimistic_estimate"):
            return
        if search_space_size > self.task.result_set_size:
            # remove selectors which have to lo of an optimistic estimate
            # selectors_map = {selector : i for i,(_, selector, _) in enumerate(s)}
            stats = []
            for _, _, cov_arr in s:
                statistics = self.task.qf.calculate_statistics(
                    cov_arr, self.task.target, self.task.data
                )
                stats.append(statistics)
                quality = self.task.qf.evaluate(
                    cov_arr, self.task.target, self.task.data, statistics
                )
                ps.add_if_required(
                    self.results, None, quality, self.task, statistics=statistics
                )
            del statistics
            to_pop = []
            min_quality = ps.minimum_required_quality(self.results, self.task)
            for i, ((_, _, cov_arr), statistics) in enumerate(zip(s, stats)):
                if (
                    not self.task.qf.optimistic_estimate(
                        cov_arr, self.task.target, self.task.data, statistics
                    )
                    > min_quality
                ):
                    to_pop.append(i)
            self.task.min_quality = np.nextafter(
                float(min_quality), self.task.min_quality
            )
            for i in reversed(to_pop):
                s.pop(i)
            self.results.clear()

    def nodes_to_cls_nodes(self, nodes):
        cls_nodes = defaultdict(list)
        for node in nodes:
            cls_nodes[node.cls].append(node)
        return cls_nodes

    def setup_from_quality_function(self, qf):
        # pylint: disable=attribute-defined-outside-init
        self.get_stats = qf.gp_get_stats
        self.get_null_vector = qf.gp_get_null_vector
        self.merge = qf.gp_merge
        self.requires_cover_arr = qf.gp_requires_cover_arr
        # pylint: enable=attribute-defined-outside-init

    def setup_constraints(self, constraints, qf):
        self.constraints_monotone = constraints
        for constraint in self.constraints_monotone:
            constraint.gp_prepare(qf)

        if self.mode == "t_d" and len(self.constraints_monotone) == 0:
            warnings.warn(
                """Poor runtime expected: Top down method does not use
                optimistic estimates and no constraints were provided""",
                UserWarning,
            )

        if len(constraints) == 1:
            self.check_constraints = constraints[0].gp_is_satisfied

    def check_constraints(self, node):  # pylint: disable=method-hidden
        return all(
            constraint.gp_is_satisfied(node) for constraint in self.constraints_monotone
        )

    def setup(self, task):
        self.task = task
        task.qf.calculate_constant_statistics(task.data, task.target)
        self.depth = task.depth
        self.setup_constraints(task.constraints_monotone, task.qf)

        self.setup_from_quality_function(task.qf)

    def create_initial_tree(self, arrs):
        # Create tree
        root = self.GP_node(-1, -1, None, {}, self.get_null_vector())
        nodes = []
        for row_index, row in self.tqdm(
            enumerate(arrs), "creating tree", total=len(arrs)
        ):
            self.normal_insert(
                root, nodes, self.get_stats(row_index), np.nonzero(row)[0]
            )
        nodes.append(root)
        return root, nodes

    def execute(self, task):
        assert self.mode in ("b_u", "t_d"), "mode needs to be either b_u or t_d"
        self.setup(task)

        selectors_sorted, arrs = self.prepare_selectors(task.search_space, task.data)
        root, nodes = self.create_initial_tree(arrs)

        # mine tree
        cls_nodes = self.nodes_to_cls_nodes(nodes)
        if self.mode == "b_u":
            self.recurse(cls_nodes, tuple())
        else:  # self.mode == "t_d"
            results = self.recurse_top_down(cls_nodes, root)
            results = self.calculate_quality_function_for_patterns(task, results, arrs)
            for quality, sg, stats in results:
                ps.add_if_required(
                    self.results, sg, quality, self.task, statistics=stats
                )
        self.results = self.convert_results_to_subgroups(self.results, selectors_sorted)

        self.results = ps.prepare_subgroup_discovery_result(self.results, task)
        return ps.SubgroupDiscoveryResult(self.results, task)

    def convert_results_to_subgroups(self, results, selectors_sorted):
        new_result = []
        for quality, indices, stats in results:
            selectors = [selectors_sorted[i] for i in indices]
            sg = ps.Conjunction(selectors)
            new_result.append((quality, sg, stats))
        return new_result

    def calculate_quality_function_for_patterns(self, task, results, arrs):
        out = []
        for indices, gp_params in self.tqdm(
            results,
            "computing quality function",
        ):
            if self.requires_cover_arr:
                if len(indices) == 1:
                    cover_arr = arrs[:, indices[0]]
                else:
                    cover_arr = np.all([arrs[:, i] for i in indices])
                statistics = task.qf.gp_get_params(cover_arr, gp_params)
                sg = cover_arr
            else:
                statistics = task.qf.gp_get_params(None, gp_params)
                sg = None
            # qual1 = task.qf.evaluate(sg, task.qf.calculate_statistics(sg, task.data))
            qual2 = task.qf.evaluate(sg, task.target, task.data, statistics)
            out.append((qual2, indices, statistics))
        return out

    def normal_insert(self, root, nodes, new_stats, classes):
        node = root
        for cls in classes:
            if cls not in node.children:
                new_child = self.GP_node(
                    cls, len(nodes), node, {}, self.get_null_vector()
                )
                nodes.append(new_child)
                node.children[cls] = new_child
            self.merge(node.stats, new_stats)
            node = node.children[cls]
        self.merge(node.stats, new_stats)
        return node

    def add_if_required(self, prefix, gp_stats):
        statistics = self.task.qf.gp_get_params(None, gp_stats)
        quality = self.task.qf.evaluate(None, None, None, statistics)
        ps.add_if_required(
            self.results, prefix, quality, self.task, statistics=statistics
        )

    def recurse(self, cls_nodes, prefix, is_single_path=False):
        if len(cls_nodes) == 0:
            raise RuntimeError  # pragma: no cover
        self.add_if_required(prefix, cls_nodes[-1][0].stats)
        if len(prefix) >= self.depth:
            return  # pragma: no cover

        stats_dict = self.get_stats_for_class(cls_nodes)
        if not self.requires_cover_arr:
            statistics = self.task.qf.gp_get_params(None, cls_nodes[-1][0].stats)
            optimistic_estimate = self.task.qf.optimistic_estimate(
                None, self.task.target, self.task.data, statistics
            )
            if not optimistic_estimate >= ps.minimum_required_quality(
                self.results, self.task
            ):
                return
        if is_single_path:
            if len(cls_nodes) == 1 and -1 in cls_nodes:
                return
            del stats_dict[-1]  # remove root node
            all_combinations = ps.powerset(
                stats_dict.keys(), max_length=self.depth - len(prefix) + 1
            )

            for comb in all_combinations:
                # it might still be, that stats_dict[comb[-1]] is wrong
                # if that is the case then
                # stats_dict[comb[0]] is correct
                if len(comb) > 0:
                    self.add_if_required(prefix + comb, stats_dict[comb[-1]])
        else:
            for cls, nodes in cls_nodes.items():
                if cls >= 0:
                    if self.check_constraints(stats_dict[cls]):
                        if len(prefix) == (self.depth - 1):
                            self.add_if_required((*prefix, cls), stats_dict[cls])
                        else:
                            is_single_path_now = len(nodes) == 1
                            new_tree = self.create_new_tree_from_nodes(nodes)
                            assert len(new_tree) > 0
                            self.recurse(new_tree, (*prefix, cls), is_single_path_now)

    def recurse_top_down(self, cls_nodes, root, depth_in=0):
        # print(f"{depth_in}"+"\t"*depth_in+str(root.cls))
        # print("init root", root.cls)
        # print(depth_in)
        # self.check_tree_is_ordered(root)
        results = []

        curr_depth = depth_in
        stats_dict = self.get_stats_for_class(cls_nodes)
        is_valid_class = {
            key: self.check_constraints(gp_stats)
            for key, gp_stats in stats_dict.items()
        }
        # init_class = root.cls
        # direct_child = None
        init_root = root
        alpha = []
        while True:
            if root.cls == -1:
                pass
            else:
                alpha.append(root.cls)
            if len(root.children) == 1 and curr_depth <= self.depth:
                # print(f"Path optmization {len(root.children)}")
                #    curr_depth += 1

                potential_root = next(iter(root.children.values()))
                if is_valid_class[potential_root.cls]:
                    root = potential_root
                else:
                    break
            else:
                break
        # self.get_prefixes_top_down(alpha,  max_length=self.depth - depth_in + 1) #
        # assert len(alpha) > 0
        prefixes = list(ps.powerset(alpha, max_length=self.depth - depth_in + 1))[1:]

        # prefixes = list(map(lambda x: sum(x, tuple()), prefixes))
        # print(root.cls, list(root.children), prefixes)
        # print("AAA", list(cls_nodes.keys()))

        if init_root.cls == -1:
            prefixes.append(tuple())
        for prefix in prefixes:
            if len(prefix) == 0:
                if is_valid_class[init_root.cls]:
                    results.append((prefix, stats_dict[init_root.cls]))
                continue
            cls = prefix[-1]
            assert is_valid_class[cls]
            results.append((prefix, stats_dict[cls]))

        # suffixes = [((), root.stats)]

        suffixes = []
        if curr_depth == (self.depth - 1):
            # print(f"{depth_in}"+"\t"*depth_in+"B")
            for cls, stats in stats_dict.items():
                if cls < 0 or cls in alpha:
                    continue
                assert cls > max(
                    alpha
                ), f"{cls} {max(alpha)}, {alpha}, {list(stats_dict.keys())}"
                suffixes.append(((cls,), stats))
        else:
            # print(f"{depth_in}"+"\t"*depth_in+"A")
            for cls in stats_dict:
                if cls < 0 or cls in alpha:
                    continue
                if is_valid_class[cls]:
                    # Future: There is also the possibility
                    # to compute the stats_dict of the prefix tree
                    # without creating the prefix tree first
                    # This might be useful if curr_depth == self.depth - 2
                    # as we need not recreate the tree
                    new_root, nodes = self.get_top_down_tree_for_class(
                        cls_nodes, cls, is_valid_class
                    )
                    # self.check_tree_is_ordered(new_root)
                    # self.check_tree_is_ordered(init_root)
                    assert len(nodes) > 0
                    new_cls_nodes = self.nodes_to_cls_nodes(nodes)
                    # new_dict = self.get_stats_for_class(new_cls_nodes)
                    # for key, value in new_dict.items():
                    #    if isinstance(stats_dict[key], dict):
                    #        continue
                    #    assert stats_dict[key][0]>=value[0], \
                    #       f"{stats_dict[key][0]} {value[0]}"
                    #    assert stats_dict[key][1]>=value[1], \
                    #       f"{stats_dict[key][1]} {value[1]}"
                    # print("  " * curr_depth, cls, curr_depth, len(new_cls_nodes))
                    suffixes.extend(
                        self.recurse_top_down(new_cls_nodes, new_root, curr_depth + 1)
                    )
        # if prefixes == [(12,), (13,)]:
        #    print(f"{depth_in}"+"\t"*depth_in+ "pre, suf", prefixes)

        # the combination below can be optimized to avoid the if
        # by first grouping them by length
        results.extend(
            [
                ((*pre, *suf), gp_stats)
                for pre, (suf, gp_stats) in itertools.product(prefixes, suffixes)
                if len(pre) + len(suf) <= self.depth
                and (len(pre) == 0 or pre[-1] < suf[0])
            ]
        )
        # if prefixes == [(12,), (13,)]:
        #    print(f"{depth_in}"+"\t"*depth_in+ "results", results)
        # print()
        return results

    def check_tree_is_ordered(self, root, prefix=None):  # pragma: no cover
        """Verify that the nodes of a tree are sorted in ascending order"""
        if prefix is None:
            prefix = []
        s = {root.cls}
        for cls, child in root.children.items():
            assert child.cls > root.cls, f"{prefix} , {cls}"
            s2 = self.check_tree_is_ordered(child, prefix + [cls])
            s = s.union(s2)
        return s

    def get_top_down_tree_for_class(self, cls_nodes, cls, is_valid_class):
        # Future: Can eventually also remove infrequent nodes already
        # during tree creation
        base_root = None
        nodes = []
        if len(cls_nodes[cls]) > 0 and is_valid_class[cls]:  # pragma: no branch okay
            base_root = self.create_copy_of_tree_top_down(
                cls_nodes[cls][0], nodes, is_valid_class=is_valid_class
            )
            for from_root in cls_nodes[cls][1:]:
                self.merge_trees_top_down(nodes, base_root, from_root, is_valid_class)
        return base_root, nodes

    def create_copy_of_tree_top_down(
        self, from_root, nodes=None, parent=None, is_valid_class=None
    ):
        if nodes is None:
            nodes = []  # pragma: no cover
        # if len(nodes) == 0:
        #    root_cls = -1
        children = {}
        new_root = self.GP_node(
            from_root.cls, len(nodes), parent, children, from_root.stats.copy()
        )
        nodes.append(new_root)
        for child_cls, child in from_root.children.items():
            if (
                is_valid_class is None or child_cls in is_valid_class
            ):  # pragma: no branch okay
                new_child = self.create_copy_of_tree_top_down(
                    child, nodes, new_root, is_valid_class=is_valid_class
                )
                children[child_cls] = new_child
        return new_root

    def merge_trees_top_down(self, nodes, mutable_root, from_root, is_valid_class):
        self.merge(mutable_root.stats, from_root.stats)
        for cls in from_root.children:
            if cls not in mutable_root.children:
                new_child = self.create_copy_of_tree_top_down(
                    from_root.children[cls],
                    nodes,
                    mutable_root,
                    is_valid_class=is_valid_class,
                )
                mutable_root.children[cls] = new_child
            else:
                self.merge_trees_top_down(
                    nodes,
                    mutable_root.children[cls],
                    from_root.children[cls],
                    is_valid_class=is_valid_class,
                )

    def get_stats_for_class(self, cls_nodes):
        out = {}
        for key, nodes in cls_nodes.items():
            s = self.get_null_vector()
            for node in nodes:
                self.merge(s, node.stats)
            out[key] = s
        return out

    def create_new_tree_from_nodes(self, nodes):
        new_nodes = {}
        for node in nodes:
            nodes_upwards = self.get_nodes_upwards(node)
            self.create_copy_of_path(nodes_upwards[1:], new_nodes, node.stats)

        # self.remove_infrequent_nodes(new_nodes)
        cls_nodes = defaultdict(list)
        for new_node in new_nodes.values():
            cls_nodes[new_node.cls].append(new_node)
        return cls_nodes

    def create_copy_of_path(self, nodes, new_nodes, stats):
        parent = None
        for node in reversed(nodes):
            if node.id not in new_nodes:
                new_node = self.GP_node(node.cls, node.id, parent, {}, stats.copy())
                new_nodes[node.id] = new_node
            else:
                new_node = new_nodes[node.id]
                self.merge(new_node.stats, stats)
            if parent is not None:
                parent.children[new_node.cls] = new_node
            parent = new_node

    def get_nodes_upwards(self, node):
        ref = node
        path = []
        while True:
            path.append(ref)
            ref = ref.parent
            if ref is None:
                break
        return path

    def to_file(self, task, path):
        self.setup(task)
        _, arrs = self.prepare_selectors(task.search_space, task.data)

        # Create tree
        to_str = task.qf.gp_to_str
        path = Path(path).absolute()
        print(path)
        with open(path, "w", encoding="utf-8") as f:
            for row_index, row in self.tqdm(
                enumerate(arrs), "creating tree", total=len(arrs)
            ):
                f.write(
                    " ".join(map(str, np.nonzero(row)[0]))
                    + " "
                    + to_str(self.get_stats(row_index))
                    + "\n"
                )



import unittest

import numpy as np
import pandas as pd

import pysubgroup as ps


class TestRepresentation(unittest.TestCase):
    def setUp(self):
        self.A = np.array([0, 0, 1, 1, 0, 0, 1, 1, 1, 1], dtype=bool)
        self.A1 = ps.EqualitySelector("columnA", True)
        self.A0 = ps.EqualitySelector("columnA", False)

        self.B = np.array(["A", "B", "C", "C", "B", "A", "D", "A", "A", "A"])
        self.BA = ps.EqualitySelector("columnB", "A")
        self.BC = ps.EqualitySelector("columnB", "C")

        self.C = np.array([np.nan, np.nan, 1.1, 1.1, 2, 2, 2, 2, 2, 2])
        self.CA = ps.EqualitySelector("columnC", 1.1)
        self.CNan = ps.EqualitySelector("columnC", np.nan)

        self.df = pd.DataFrame.from_dict(
            {"columnA": self.A, "columnB": self.B, "columnC": self.C}
        )

    def test_BitSet(self):
        with ps.BitSetRepresentation(
            self.df, [self.A1, self.A0, self.BA, self.BC, self.CA, self.CNan]
        ) as representation:
            # pylint: disable=no-member
            np.testing.assert_array_equal(self.A1.representation, self.A)
            np.testing.assert_array_equal(
                self.A0.representation, np.logical_not(self.A)
            )

            np.testing.assert_array_equal(
                self.BA.representation, [1, 0, 0, 0, 0, 1, 0, 1, 1, 1]
            )
            np.testing.assert_array_equal(
                self.BC.representation, [0, 0, 1, 1, 0, 0, 0, 0, 0, 0]
            )

            np.testing.assert_array_equal(
                self.CA.representation, [0, 0, 1, 1, 0, 0, 0, 0, 0, 0]
            )
            np.testing.assert_array_equal(
                self.CNan.representation, [1, 1, 0, 0, 0, 0, 0, 0, 0, 0]
            )

            np.testing.assert_array_equal(
                representation.Conjunction([self.BA, self.CNan]).representation,
                [1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            )
            np.testing.assert_array_equal(
                representation.Disjunction([self.BA, self.BC]).representation,
                [1, 0, 1, 1, 0, 1, 0, 1, 1, 1],
            )

            np.testing.assert_array_equal(
                representation.Disjunction().representation,
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            )
            tmp_disjunction = representation.Disjunction(self.BA)
            tmp_disjunction.append_or(self.BC)
            np.testing.assert_array_equal(
                tmp_disjunction.representation, [1, 0, 1, 1, 0, 1, 0, 1, 1, 1]
            )
            # pylint: enable=no-member

    def test_Set(self):
        with ps.SetRepresentation(
            self.df, [self.A1, self.A0, self.BA, self.BC, self.CA, self.CNan]
        ) as representation:
            self.assertEqual(
                self.A1.representation, {2, 3, 6, 7, 8, 9}
            )  # pylint: disable=no-member
            self.assertEqual(
                self.A0.representation, {0, 1, 4, 5}
            )  # pylint: disable=no-member

            self.assertEqual(
                self.BA.representation, {0, 5, 7, 8, 9}
            )  # pylint: disable=no-member
            self.assertEqual(
                self.BC.representation, {2, 3}
            )  # pylint: disable=no-member

            self.assertEqual(
                self.CA.representation, {2, 3}
            )  # pylint: disable=no-member
            self.assertEqual(
                self.CNan.representation, {0, 1}
            )  # pylint: disable=no-member

            self.assertEqual(
                representation.Conjunction([self.BA, self.CNan]).representation, {0}
            )  # pylint: disable=no-member
            self.assertEqual(
                representation.Conjunction([self.A0, self.CNan]).representation, {0, 1}
            )  # pylint: disable=no-member

    def test_NumpySet(self):
        with ps.NumpySetRepresentation(
            self.df, [self.A1, self.A0, self.BA, self.BC, self.CA, self.CNan]
        ) as representation:
            np.testing.assert_array_equal(
                self.A1.representation, [2, 3, 6, 7, 8, 9]
            )  # pylint: disable=no-member
            np.testing.assert_array_equal(
                self.A0.representation, [0, 1, 4, 5]
            )  # pylint: disable=no-member

            np.testing.assert_array_equal(
                self.BA.representation, [0, 5, 7, 8, 9]
            )  # pylint: disable=no-member
            np.testing.assert_array_equal(
                self.BC.representation, [2, 3]
            )  # pylint: disable=no-member

            np.testing.assert_array_equal(
                self.CA.representation, [2, 3]
            )  # pylint: disable=no-member
            np.testing.assert_array_equal(
                self.CNan.representation, [0, 1]
            )  # pylint: disable=no-member

            np.testing.assert_array_equal(
                representation.Conjunction([self.BA, self.CNan]).representation, [0]
            )  # pylint: disable=no-member
            np.testing.assert_array_equal(
                representation.Conjunction([self.A0, self.CNan]).representation, [0, 1]
            )  # pylint: disable=no-member


if __name__ == "__main__":
    unittest.main()



import numpy as np

from pysubgroup.subgroup_description import Conjunction, Disjunction


class RepresentationBase:
    def __init__(self, new_conjunction, selectors_to_patch):
        self._new_conjunction = new_conjunction
        self.previous_conjunction = None
        self.selectors_to_patch = selectors_to_patch

    def patch_all_selectors(self):
        for sel in self.selectors_to_patch:
            self.patch_selector(sel)

    def patch_selector(self, sel):  # pragma: no cover
        raise NotImplementedError()  # pragma: no cover

    def patch_classes(self):
        pass

    def undo_patch_classes(self):
        pass

    def __enter__(self):
        self.patch_classes()
        self.patch_all_selectors()
        return self

    def __exit__(self, *args):
        self.undo_patch_classes()


class BitSet_Conjunction(Conjunction):
    n_instances = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.representation = self.compute_representation()

    def compute_representation(self):
        # empty description ==> return a list of all '1's
        if not self._selectors:
            return np.full(BitSet_Conjunction.n_instances, True, dtype=bool)
        # non-empty description
        return np.all([sel.representation for sel in self._selectors], axis=0)

    @property
    def size_sg(self):
        return np.count_nonzero(self.representation)

    def append_and(self, to_append):
        super().append_and(to_append)
        self.representation = np.logical_and(
            self.representation, to_append.representation
        )

    @property
    def __array_interface__(self):
        return self.representation.__array_interface__


class BitSet_Disjunction(Disjunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.representation = self.compute_representation()

    def compute_representation(self):
        # empty description ==> return a list of all '1's
        if not self._selectors:
            return np.full(BitSet_Conjunction.n_instances, False, dtype=bool)
        # non-empty description
        return np.any([sel.representation for sel in self._selectors], axis=0)

    @property
    def size_sg(self):
        return np.count_nonzero(self.representation)

    def append_or(self, to_append):
        super().append_or(to_append)
        self.representation = np.logical_or(
            self.representation, to_append.representation
        )

    @property
    def __array_interface__(self):
        return self.representation.__array_interface__


class BitSetRepresentation(RepresentationBase):
    Conjunction = BitSet_Conjunction
    Disjunction = BitSet_Disjunction

    def __init__(self, df, selectors_to_patch):
        self.df = df
        super().__init__(BitSet_Conjunction, selectors_to_patch)

    def patch_selector(self, sel):
        sel.representation = sel.covers(self.df)
        sel.size_sg = np.count_nonzero(sel.representation)

    def patch_classes(self):
        BitSet_Conjunction.n_instances = len(self.df)
        super().patch_classes()


class Set_Conjunction(Conjunction):
    all_set = set()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.representation = self.compute_representation()
        self.arr_for_interface = np.array(list(self.representation), dtype=int)

    def compute_representation(self):
        # empty description ==> return a list of all '1's
        if not self._selectors:
            return Set_Conjunction.all_set
        # non-empty description
        return set.intersection(*[sel.representation for sel in self._selectors])

    @property
    def size_sg(self):
        return len(self.representation)

    def append_and(self, to_append):
        super().append_and(to_append)
        self.representation = self.representation.intersection(to_append.representation)
        self.arr_for_interface = np.array(list(self.representation), dtype=int)

    @property
    def __array_interface__(self):
        return self.arr_for_interface.__array_interface__  # pylint: disable=no-member


class SetRepresentation(RepresentationBase):
    Conjunction = Set_Conjunction

    def __init__(self, df, selectors_to_patch):
        self.df = df
        super().__init__(Set_Conjunction, selectors_to_patch)

    def patch_selector(self, sel):
        sel.representation = set(*np.nonzero(sel.covers(self.df)))
        sel.size_sg = len(sel.representation)

    def patch_classes(self):
        Set_Conjunction.all_set = set(self.df.index)
        super().patch_classes()


class NumpySet_Conjunction(Conjunction):
    all_set = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.representation = self.compute_representation()

    def compute_representation(self):
        # empty description ==> return a list of all '1's
        if not self._selectors:
            return NumpySet_Conjunction.all_set
        start = self._selectors[0].representation
        for sel in self._selectors[1:]:
            start = np.intersect1d(start, sel.representation, assume_unique=True)
        return start

    @property
    def size_sg(self):
        return len(self.representation)

    def append_and(self, to_append):
        super().append_and(to_append)
        # self._selectors.append(to_append)
        self.representation = np.intersect1d(
            self.representation, to_append.representation, True
        )

    @property
    def __array_interface__(self):
        return self.representation.__array_interface__


class NumpySetRepresentation(RepresentationBase):
    Conjunction = NumpySet_Conjunction

    def __init__(self, df, selectors_to_patch):
        self.df = df
        super().__init__(NumpySet_Conjunction, selectors_to_patch)

    def patch_selector(self, sel):
        sel.representation = np.nonzero(sel.covers(self.df))[0]
        sel.size_sg = len(sel.representation)

    def patch_classes(self):
        NumpySet_Conjunction.all_set = np.arange(len(self.df))
        super().patch_classes()



# flake8: noqa: E501

import unittest
from copy import copy

from algorithms_testing import TestAlgorithmsBase
from t_utils import conjunctions_from_str

import pysubgroup as ps
from pysubgroup.datasets import get_credit_data


class TestStandardQFNumeric(unittest.TestCase):
    def test_constructor(self):
        ps.StandardQFNumeric(0)
        ps.StandardQFNumeric(1.0)
        ps.StandardQFNumeric(0, invert=True)
        ps.StandardQFNumeric(0, invert=False)

        with self.assertRaises(ValueError):
            ps.StandardQFNumeric("test")

        ps.StandardQFNumeric(0, estimator="sum")
        ps.StandardQFNumeric(0, estimator="average")
        ps.StandardQFNumeric(0, estimator="order")

        with self.assertRaises(ValueError):
            ps.StandardQFNumeric(0, estimator="bla")


class TestStandardQFNumericMedian(TestAlgorithmsBase, unittest.TestCase):
    def test_constructor(self):
        ps.StandardQFNumeric(0, centroid="median")
        ps.StandardQFNumeric(1.0, centroid="median")
        ps.StandardQFNumeric(0, invert=True, centroid="median")
        ps.StandardQFNumeric(0, invert=False, centroid="median")

        ps.StandardQFNumeric(0, centroid="sorted_median")

        with self.assertRaises(ValueError):
            ps.StandardQFNumeric("test", centroid="median")

        with self.assertRaises(ValueError):
            ps.StandardQFNumeric(0, centroid="bla")

        ps.StandardQFNumeric(0, estimator="max", centroid="median")
        # ps.StandardQFNumeric(0, estimator="order", centroid='median')

        with self.assertRaises(AssertionError):
            ps.StandardQFNumeric(0, estimator="sum", centroid="median")
        with self.assertRaises(AssertionError):
            ps.StandardQFNumeric(0, estimator="average", centroid="median")

        with self.assertRaises(AssertionError):
            ps.StandardQFNumeric(0, estimator="bla", centroid="median")

    def setUp(self):
        l = conjunctions_from_str(
            """316646.0    job=='b'high qualif/self emp/mgmt''
   310615.0    foreign_worker=='b'yes'' AND job=='b'high qualif/self emp/mgmt''
   297844.5    foreign_worker=='b'yes'' AND own_telephone=='b'yes'' AND property_magnitude=='b'no known property''
   297844.5    own_telephone=='b'yes'' AND property_magnitude=='b'no known property''
   288480.5    job=='b'high qualif/self emp/mgmt'' AND own_telephone=='b'yes''
   283002.0    own_telephone=='b'yes''
   282217.5    class=='b'bad'' AND own_telephone=='b'yes''
   282000.0    other_parties=='b'none'' AND own_telephone=='b'yes''
   276312.5    foreign_worker=='b'yes'' AND job=='b'high qualif/self emp/mgmt'' AND own_telephone=='b'yes''
   275467.5    foreign_worker=='b'yes'' AND other_parties=='b'none'' AND own_telephone=='b'yes'' """
        )
        self.result = [conjunction for quality, conjunction in l]
        self.qualities = [quality for quality, conjunction in l]

        data = get_credit_data()
        target = ps.NumericTarget("credit_amount")
        searchSpace_Nominal = ps.create_nominal_selectors(
            data, ignore=["credit_amount"]
        )
        searchSpace_Numeric = (
            []
        )  # ps.create_numeric_selectors(data, ignore=['credit_amount'], nbins=10)
        searchSpace = searchSpace_Nominal + searchSpace_Numeric
        self.task = ps.SubgroupDiscoveryTask(
            data,
            target,
            searchSpace,
            result_set_size=10,
            depth=5,
            qf=ps.StandardQFNumeric(1, False, estimator="max", centroid="median"),
        )

    def Atest_Apriori_no_numba(self):
        algorithm = ps.Apriori(use_numba=False)
        algorithm.use_vectorization = False
        self.runAlgorithm(
            algorithm,
            "Median Apriori, use_numba=False",
            self.result,
            self.qualities,
            self.task,
        )

    def test_Apriori_no_numba_sorted_median(self):
        algorithm = ps.Apriori(use_numba=False)
        algorithm.use_vectorization = False

        qf = ps.StandardQFNumeric(1, False, estimator="max", centroid="sorted_median")
        df = self.task.data
        df = df.sort_values(by=self.task.target.target_variable)
        task = ps.SubgroupDiscoveryTask(
            df,
            self.task.target,
            self.task.search_space,
            qf,
            result_set_size=10,
            depth=5,
        )
        self.runAlgorithm(
            algorithm,
            "Quick Median Apriori, use_numba=False",
            self.result,
            self.qualities,
            task,
        )


#    def test_SimpleDFS(self): # terribly slow
#        self.runAlgorithm(
#            ps.SimpleDFS(), "Median SimpleDFS", self.result, self.qualities, self.task
#        )


class TestNumericTarget(unittest.TestCase):
    def setUp(self):
        records = [(1, 20), (1, 10), (1, 10), (1, 0), (0, -20), (0, -10), (0, -10)]
        self.df = pd.DataFrame.from_records(records, columns=("A", "target"))

    def test_get_base_statistics(self):
        sg = ps.EqualitySelector("A", 1)
        target = ps.NumericTarget("target")
        self.assertEqual(target.get_base_statistics(sg, self.df), (7, 0.0, 4, 10.0))

    def test_calculate_statistics(self):
        target = ps.NumericTarget("target")
        sg = ps.EqualitySelector("A", 1)
        statistics = target.calculate_statistics(sg, self.df, None)
        statistics2 = target.calculate_statistics(
            sg, self.df, cached_statistics=statistics
        )
        self.assertIs(statistics, statistics2)
        del statistics["size_sg"]
        statistics3 = target.calculate_statistics(
            sg, self.df, cached_statistics=statistics
        )


class TestStandardQFNumericTscore(unittest.TestCase):
    def test_basics(self):
        epsilon = 0.001
        records = [
            (1, 20),
            (1, 10),
            (1, 10 + epsilon),
            (1, 0),
            (0, -20),
            (0, -10),
            (0, -10 - epsilon),
        ]
        df = pd.DataFrame.from_records(records, columns=("A", "target"))
        target = ps.NumericTarget("target")
        sg = ps.EqualitySelector("A", 1)
        qf = ps.StandardQFNumericTscore()
        qf.evaluate(sg, target, df, None)
        qf.optimistic_estimate(sg, target, df, None)

        # test empty subgroup
        sg0 = ps.EqualitySelector("A", 2)
        qf.evaluate(sg0, target, df, None)
        qf.optimistic_estimate(sg0, target, df, None)


class TestAlgorithmsWithNumericTarget(TestAlgorithmsBase, unittest.TestCase):
    def setUp(self):
        NS_telephone = ps.EqualitySelector("own_telephone", b"yes")
        NS_foreign_worker = ps.EqualitySelector("foreign_worker", b"yes")
        NS_other_parties = ps.EqualitySelector("other_parties", b"none")
        NS_personal = ps.EqualitySelector("personal_status", b"male single")
        NS_job = ps.EqualitySelector("job", b"high qualif/self emp/mgmt")
        NS_class = ps.EqualitySelector("class", b"bad")

        o = [
            [NS_telephone],
            [NS_foreign_worker, NS_telephone],
            [NS_other_parties, NS_telephone],
            [NS_foreign_worker, NS_telephone, NS_personal],
            [NS_telephone, NS_personal],
            [NS_foreign_worker, NS_other_parties, NS_telephone],
            [NS_job],
            [NS_class, NS_telephone],
            [NS_foreign_worker, NS_job],
            [NS_foreign_worker, NS_other_parties, NS_telephone, NS_personal],
        ]
        self.result = list(map(ps.Conjunction, o))
        self.qualities = [
            383476.7679999999,
            361710.05800000014,
            345352.9920000001,
            338205.08,
            336857.8220000001,
            323586.28200000006,
            320306.81600000005,
            300963.84599999996,
            299447.332,
            297422.98200000013,
        ]

        data = get_credit_data()
        target = ps.NumericTarget("credit_amount")
        searchSpace_Nominal = ps.create_nominal_selectors(
            data, ignore=["credit_amount"]
        )
        searchSpace_Numeric = (
            []
        )  # ps.create_numeric_selectors(data, ignore=['credit_amount'], nbins=10)
        searchSpace = searchSpace_Nominal + searchSpace_Numeric
        self.task = ps.SubgroupDiscoveryTask(
            data,
            target,
            searchSpace,
            result_set_size=10,
            depth=5,
            qf=ps.StandardQFNumeric(1, False, "sum"),
        )

    def test_SimpleDFS(self):
        self.runAlgorithm(
            ps.SimpleDFS(), "SimpleDFS", self.result, self.qualities, self.task
        )

    def test_DFS_average(self):
        self.task.qf = ps.StandardQFNumeric(self.task.qf.a, False, "average")
        self.runAlgorithm(
            ps.DFS(ps.BitSetRepresentation),
            "DFS average",
            self.result,
            self.qualities,
            self.task,
        )

    def test_DFS_order_with_numba(self):
        try:
            import numba
        except ImportError:
            self.skipTest("No numba installed")
        self.task.qf = ps.StandardQFNumeric(self.task.qf.a, False, "order")
        self.runAlgorithm(
            ps.DFS(ps.BitSetRepresentation),
            "DFS order with numba",
            self.result,
            self.qualities,
            self.task,
        )

    def test_DFS_order_no_numba(self):
        self.task.qf = ps.StandardQFNumeric(self.task.qf.a, False, "order")
        self.task.qf.estimator.use_numba = False
        self.runAlgorithm(
            ps.DFS(ps.BitSetRepresentation),
            "DFS order no numba",
            self.result,
            self.qualities,
            self.task,
        )

    def test_DFS_sum(self):
        self.task.qf = ps.StandardQFNumeric(self.task.qf.a, False, "sum")
        self.runAlgorithm(
            ps.DFS(ps.BitSetRepresentation),
            "DFS sum",
            self.result,
            self.qualities,
            self.task,
        )

    def test_BeamSearch_sum(self):
        self.task.qf = ps.StandardQFNumeric(self.task.qf.a, False, "sum")
        self.runAlgorithm(
            ps.BeamSearch(beam_width=self.task.result_set_size),
            "BeamSearch sum",
            self.result,
            self.qualities,
            self.task,
        )

    def test_BeamSearch_average(self):
        self.task.qf = ps.StandardQFNumeric(self.task.qf.a, False, "average")
        self.runAlgorithm(
            ps.BeamSearch(beam_width=self.task.result_set_size),
            "BeamSearch average",
            self.result,
            self.qualities,
            self.task,
        )

    def test_BeamSearch_order(self):
        self.task.qf = ps.StandardQFNumeric(self.task.qf.a, False, "order")
        self.runAlgorithm(
            ps.BeamSearch(beam_width=self.task.result_set_size),
            "BeamSearch order",
            self.result,
            self.qualities,
            self.task,
        )

    def test_Apriori_no_numba(self):
        self.runAlgorithm(
            ps.Apriori(use_numba=False),
            "Apriori use_numba=False",
            self.result,
            self.qualities,
            self.task,
        )

    def test_Apriori_with_numba(self):
        self.runAlgorithm(
            ps.Apriori(use_numba=True),
            "Apriori use_numba=True",
            self.result,
            self.qualities,
            self.task,
        )

    def test_DFSNumeric(self):
        self.runAlgorithm(
            ps.DFSNumeric(), "DFS_numeric", self.result, self.qualities, self.task
        )
        # print('   Number of call to qf:', algo.num_calls)

    # def test_SimpleSearch(self):
    #   self.runAlgorithm(ps.SimpleSearch(), "SimpleSearch", self.result, self.qualities, self.task)


import pandas as pd


class TestNumericEstimators(unittest.TestCase):
    def test_estimator1(self):
        records = [(1, 100), (1, 75), (1, 53), (1, 12), (0, 11), (0, 49)]
        df = pd.DataFrame.from_records(records, columns=["A", "Target"])
        T = ps.NumericTarget("Target")
        sel = ps.EqualitySelector("A", 1)

        qf = ps.StandardQFNumeric(a=1)
        self.assertEqual(qf.optimistic_estimate(sel, T, df), 78)

        for a in [1, 0.5, 0]:
            qf = ps.StandardQFNumeric(a=a, estimator="max")
            self.assertEqual(qf.optimistic_estimate(sel, T, df), 3**a * 50)


if __name__ == "__main__":
    unittest.main()


# 383476.7679999999:      own_telephone=='b'yes''
# 361710.05800000014:     foreign_worker=='b'yes'' and own_telephone=='b'yes''
# 345352.9920000001:      other_parties=='b'none'' and own_telephone=='b'yes''
# 338205.08:      foreign_worker=='b'yes'' and own_telephone=='b'yes'' and personal_status=='b'male single''
# 336857.8220000001:      own_telephone=='b'yes'' and personal_status=='b'male single''
# 323586.28200000006:     foreign_worker=='b'yes'' and other_parties=='b'none'' and own_telephone=='b'yes''
# 320306.81600000005:     job=='b'high qualif/self emp/mgmt''
# 300963.84599999996:     class=='b'bad'' and own_telephone=='b'yes''
# 299447.332:     foreign_worker=='b'yes'' and job=='b'high qualif/self emp/mgmt''
# 297422.98200000013:     foreign_worker=='b'yes'' and other_parties=='b'none'' and own_telephone=='b'yes'' and personal_status=='b'male single''

# 639577.0460000001:   duration>=30.0
# 624424.3040000001:   duration>=30.0 AND foreign_worker=='b'yes''
# 579219.206:  duration>=30.0 AND other_parties=='b'none''
# 564066.4640000002:   duration>=30.0 AND foreign_worker=='b'yes'' AND other_parties=='b'none''
# 547252.302:  duration>=30.0 AND num_dependents==1.0
# 532099.56:   duration>=30.0 AND foreign_worker=='b'yes'' AND num_dependents==1.0
# 491104.688:  duration>=30.0 AND num_dependents==1.0 AND other_parties=='b'none''
# 490633.1400000001:   duration>=30.0 AND foreign_worker=='b'yes'' AND other_payment_plans=='b'none''
# 490633.1400000001:   duration>=30.0 AND other_payment_plans=='b'none''
# 475951.94600000005:  duration>=30.0 AND foreign_worker=='b'yes'' AND num_dependents==1.0 AND other_parties=='b'none''



import unittest

import pysubgroup as ps
from pysubgroup.datasets import get_credit_data


class TestAlgorithms(unittest.TestCase):
    def setUp(self) -> None:
        data = get_credit_data()
        target = ps.BinaryTarget("class", b"bad")
        searchSpace_Nominal = ps.create_nominal_selectors(data, ignore=["class"])
        searchSpace_Numeric = ps.create_numeric_selectors(data, ignore=["class"])
        searchSpace = searchSpace_Nominal + searchSpace_Numeric
        self.task = ps.SubgroupDiscoveryTask(
            data, target, searchSpace, result_set_size=2, depth=1, qf=ps.StandardQF(0.5)
        )

    def test_BeamSearch_raises(self):
        with self.assertRaises(RuntimeError):
            ps.BeamSearch(1, False).execute(self.task)

    def test_DFSNUmeric_raises(self):
        with self.assertRaises(RuntimeError):
            ps.DFSNumeric().execute(self.task)

    def test_BeamSearch_adaptive(self):
        ps.BeamSearch(1, beam_width_adaptive=True).execute(self.task)

    def test_Apriori_default_representation(self):
        ps.Apriori()

    def test_Apriori_set_representation(self):
        ps.Apriori(ps.SetRepresentation)


if __name__ == "__main__":
    unittest.main()



from collections import namedtuple

import numpy as np

import pysubgroup as ps

beta_tuple = namedtuple("beta_tuple", ["beta", "size_sg"])


class EMM_Likelihood(ps.AbstractInterestingnessMeasure):
    tpl = namedtuple(
        "EMM_Likelihood",
        ["model_params", "subgroup_likelihood", "inverse_likelihood", "size"],
    )

    def __init__(self, model):
        self.model = model
        self.has_constant_statistics = False
        self.required_stat_attrs = EMM_Likelihood.tpl._fields
        self.data_size = None

    def calculate_constant_statistics(self, data, target):
        self.model.calculate_constant_statistics(data, target)
        self.data_size = len(data)
        self.has_constant_statistics = True

    def calculate_statistics(
        self, subgroup, target, data, statistics=None
    ):  # pylint: disable=unused-argument
        cover_arr, sg_size = ps.get_cover_array_and_size(subgroup, self.data_size, data)
        params = self.model.fit(cover_arr, data)
        return self.get_tuple(sg_size, params, cover_arr)

    def get_tuple(self, sg_size, params, cover_arr):
        # numeric stability?
        all_likelihood = self.model.likelihood(
            params, np.ones(self.data_size, dtype=bool)
        )
        sg_likelihood_sum = np.sum(all_likelihood[cover_arr])
        total_likelihood_sum = np.sum(all_likelihood)
        dataset_average = np.nan
        if (self.data_size - sg_size) > 0:
            dataset_average = (total_likelihood_sum - sg_likelihood_sum) / (
                self.data_size - sg_size
            )
        sg_average = np.nan
        if sg_size > 0:
            sg_average = sg_likelihood_sum / sg_size
        return EMM_Likelihood.tpl(params, sg_average, dataset_average, sg_size)

    def evaluate(self, subgroup, target, data, statistics=None):
        statistics = self.ensure_statistics(subgroup, target, data, statistics)
        # numeric stability?
        return statistics.subgroup_likelihood - statistics.inverse_likelihood

    def gp_get_params(self, cover_arr, v):
        params = self.model.gp_get_params(v)
        sg_size = params.size_sg
        return self.get_tuple(sg_size, params, cover_arr)

    @property
    def gp_requires_cover_arr(self):
        return True

    def __getattr__(self, name):
        return getattr(self.model, name)


class PolyRegression_ModelClass:
    def __init__(self, x_name="x", y_name="y", degree=1):
        self.x_name = x_name
        self.y_name = y_name
        if degree != 1:
            raise ValueError("Currently only degree == 1 is supported")
        self.degree = degree
        self.x = None
        self.y = None
        self.has_constant_statistics = True
        super().__init__()

    def calculate_constant_statistics(
        self, data, target
    ):  # pylint: disable=unused-argument
        self.x = data[self.x_name].to_numpy()
        self.y = data[self.y_name].to_numpy()
        self.has_constant_statistics = True

    @staticmethod
    def gp_merge(u, v):
        v0 = v[0]
        u0 = u[0]
        if v0 == 0 or u0 == 0:
            d = 0
        else:
            d = v0 * u0 / (v0 + u0) * (v[1] / v0 - u[1] / u0) * (v[2] / v0 - u[2] / u0)
        u += v
        u[3] += d

    def gp_get_null_vector(self):
        return np.zeros(5)

    def gp_get_stats(self, row_index):
        x = self.x[row_index]
        return np.array([1, x, self.y[row_index], 0, x * x])

    def gp_get_params(self, v):
        size = v[0]
        if size < self.degree:
            return beta_tuple(np.full(self.degree + 1, np.nan), size)
        v1 = v[1]
        slope = v[0] * v[3] / (v[0] * v[4] - v1 * v1)
        intersept = v[2] / v[0] - slope * v[1] / v[0]
        return beta_tuple(np.array([slope, intersept]), v[0])

    def gp_to_str(self, stats):
        return " ".join(map(str, stats))

    def gp_size_sg(self, stats):
        return stats[0]

    @property
    def gp_requires_cover_arr(self):
        return False

    def fit(self, subgroup, data=None):
        cover_arr, size = ps.get_cover_array_and_size(subgroup, len(self.x), data)
        if size <= self.degree + 1:
            return beta_tuple(np.full(self.degree + 1, np.nan), size)
        return beta_tuple(
            np.polyfit(self.x[cover_arr], self.y[cover_arr], deg=self.degree), size
        )

    def likelihood(self, stats, sg):
        from scipy.stats import norm  # pylint: disable=import-outside-toplevel

        if any(np.isnan(stats.beta)):
            return np.full(self.x[sg].shape, np.nan)
        return norm.pdf(np.polyval(stats.beta, self.x[sg]) - self.y[sg])

    def loglikelihood(self, stats, sg):
        from scipy.stats import norm  # pylint: disable=import-outside-toplevel

        return norm.logpdf(np.polyval(stats.beta, self.x[sg]) - self.y[sg])



from timeit import default_timer as timer

import pysubgroup as ps
from pysubgroup.datasets import get_credit_data

data = get_credit_data()

print("running")
target = ps.BinaryTarget("class", b"bad")
search_space = ps.create_selectors(data, ignore=["class"])
task = ps.SubgroupDiscoveryTask(
    data,
    target,
    search_space,
    result_set_size=10,
    depth=3,
    qf=ps.ChiSquaredQF(direction="bidirect"),
)

start = timer()
result = ps.SimpleDFS().execute(task)
end = timer()

print("Time elapsed: ", (end - start))

for q, sg in result:
    print(str(q) + ":\t" + str(sg.subgroup_description))

# print WRAccQF().evaluate_from_dataset(data, Subgroup(target, []))



import unittest

import numpy as np
from algorithms_testing import TestAlgorithmsBase

import pysubgroup as ps
from pysubgroup.datasets import get_titanic_data

# data=get_titanic_data()
# search_space = ps.create_selectors(data, ignore="survived")
# dt = data.dtypes

# task = ps.SubgroupDiscoveryTask(
#     data, ps.FITarget, search_space, result_set_size=10, depth=5, qf=ps.CountQF())
# result = ps.SimpleDFS().execute(task)

# for (q, sg) in result:
#    print(str(q) + ":\t" + str(sg.subgroup_description))

# task = ps.SubgroupDiscoveryTask(
#     data, ps.FITarget, search_space, result_set_size=10, depth=2, qf=ps.AreaQF())
# result = ps.SimpleDFS().execute(task)

# for (q, sg) in result:
#    print(str(q) + ":\t" + str(sg.subgroup_description))

# df_test1 = pd.DataFrame.from_records(
#   [[1,1,1],[0,0,1],[0,0,1]], columns=["A", "B", "C"])


class TestFITarget(TestAlgorithmsBase, unittest.TestCase):
    def test_fi_target(self):
        target = ps.FITarget()
        self.assertEqual(target.get_attributes(), [])

        df = get_titanic_data()
        NS_embarked = ps.EqualitySelector("Embarked", "S")
        self.assertEqual(target.get_base_statistics(NS_embarked, df), 110)

        statistics = target.calculate_statistics(NS_embarked, df)
        self.assertDictEqual(statistics, {"size_sg": 110, "size_dataset": 156})
        statistics2 = target.calculate_statistics(
            NS_embarked, df, cached_statistics=statistics
        )
        self.assertIs(statistics2, statistics)

        self.assertTrue(ps.FITarget() == ps.FITarget())


class TestCountQF(TestAlgorithmsBase, unittest.TestCase):
    def test_Apriori(self):
        self.runAlgorithm(
            ps.Apriori(), "Apriori", self.result[1:], self.qualities[1:], self.task
        )

    def test_DFS(self):
        self.runAlgorithm(
            ps.SimpleDFS(), "DFS", self.result[:-1], self.qualities[:-1], self.task
        )

    def test_gp_growth(self):
        self.task.constraints_monotone.append(ps.MinSupportConstraint(20))
        self.runAlgorithm(
            ps.GpGrowth(), "GpGrowth", self.result[:-1], self.qualities[:-1], self.task
        )
        self.task.constraints_monotone.pop(-1)

    def test_gp_growth_top_down(self):
        self.task.constraints_monotone.append(ps.MinSupportConstraint(20))
        self.runAlgorithm(
            ps.GpGrowth(mode="t_d"),
            "GpGrowth",
            self.result[:-1],
            self.qualities[:-1],
            self.task,
        )
        self.task.constraints_monotone.pop(-1)

    def setUp(self):
        NS_cabin = ps.EqualitySelector("Cabin", np.nan)
        NS_embarked = ps.EqualitySelector("Embarked", "S")
        NS_male = ps.EqualitySelector("Sex", "male")
        NS_female = ps.EqualitySelector("Sex", "female")
        # NS_other_parties = ps.EqualitySelector("other_parties", b"none")
        # NS_savings_status = ps.EqualitySelector("savings_status", b"<100")
        # NS_job = ps.EqualitySelector("job", b"skilled")
        self.result = [
            ps.Conjunction([]),
            ps.Conjunction([NS_cabin]),
            ps.Conjunction([NS_embarked]),
            ps.Conjunction([NS_male]),
            ps.Conjunction([NS_cabin, NS_embarked]),
            ps.Conjunction([NS_cabin, NS_male]),
            ps.Conjunction([NS_embarked, NS_male]),
            ps.Conjunction([NS_cabin, NS_embarked, NS_male]),
            ps.Conjunction([NS_female]),
            ps.Conjunction([NS_cabin, NS_female]),
            ps.Conjunction([NS_embarked, NS_female]),
        ]

        self.qualities = [156, 125, 110, 100, 89, 82, 73, 60, 56, 43, 37]

        data = get_titanic_data()
        self.qualities2 = [np.count_nonzero(conj.covers(data)) for conj in self.result]
        self.assertEqual(self.qualities, self.qualities2)
        searchSpace = ps.create_nominal_selectors(data)
        self.task = ps.SubgroupDiscoveryTask(
            data,
            ps.FITarget(),
            searchSpace,
            result_set_size=10,
            depth=5,
            qf=ps.CountQF(),
        )


class TestAreaQF(TestAlgorithmsBase, unittest.TestCase):
    def test_SimpleSearch(self):
        self.runAlgorithm(
            ps.SimpleSearch(), "SimpleSearch", self.result, self.qualities, self.task
        )

    def test_SimpleSearch2(self):
        self.runAlgorithm(
            ps.SimpleSearch(show_progress=False),
            "SimpleSearch2",
            self.result,
            self.qualities,
            self.task,
        )

    def test_BestFirstSearch(self):
        self.runAlgorithm(
            ps.BestFirstSearch(), "SimpleSearch", self.result, self.qualities, self.task
        )

    # 178: Cabin.isnull() AND Embarked=='S'
    # 164: Cabin.isnull() AND Sex=='male'
    # 146: Embarked=='S' AND Sex=='male'
    # 125: Cabin.isnull()
    # 110: Embarked=='S'
    # 100: Sex=='male'
    # 86:  Cabin.isnull() AND Sex=='female'
    # 74:  Embarked=='S' AND Sex=='female'
    # 56:  Sex=='female'
    # 46:  Cabin.isnull() AND Embarked=='C'
    def setUp(self):
        NS_cabin = ps.EqualitySelector("Cabin", np.nan)
        NS_embarked = ps.EqualitySelector("Embarked", "S")
        NS_embarked2 = ps.EqualitySelector("Embarked", "C")
        NS_male = ps.EqualitySelector("Sex", "male")
        NS_female = ps.EqualitySelector("Sex", "female")
        # NS_other_parties = ps.EqualitySelector("other_parties", b"none")
        # NS_savings_status = ps.EqualitySelector("savings_status", b"<100")
        # NS_job = ps.EqualitySelector("job", b"skilled")
        self.result = [
            ps.Conjunction([NS_cabin, NS_embarked]),
            ps.Conjunction([NS_cabin, NS_male]),
            ps.Conjunction([NS_embarked, NS_male]),
            ps.Conjunction([NS_cabin]),
            ps.Conjunction([NS_embarked]),
            ps.Conjunction([NS_male]),
            ps.Conjunction([NS_cabin, NS_female]),
            ps.Conjunction([NS_embarked, NS_female]),
            ps.Conjunction([NS_female]),
            ps.Conjunction([NS_cabin, NS_embarked2]),
        ]

        self.qualities = [178, 164, 146, 125, 110, 100, 86, 74, 56, 46]

        data = get_titanic_data()
        self.qualities2 = [
            np.count_nonzero(conj.covers(data)) * conj.depth for conj in self.result
        ]
        self.assertEqual(self.qualities, self.qualities2)
        searchSpace = ps.create_nominal_selectors(data)
        self.task = ps.SubgroupDiscoveryTask(
            data,
            ps.FITarget(),
            searchSpace,
            result_set_size=10,
            depth=2,
            qf=ps.AreaQF(),
        )


#  156: True
#  125: Cabin.isnull()
#  110: Embarked=='S'
#  100: Sex=='male'
#  89:  Cabin.isnull() AND Embarked=='S'
#  82:  Cabin.isnull() AND Sex=='male'
#  73:  Embarked=='S' AND Sex=='male'
#  60:  Cabin.isnull() AND Embarked=='S' AND Sex=='male'
#  56:  Sex=='female'
#  43:  Cabin.isnull() AND Sex=='female'
#  37:  Embarked=='S' AND Sex=='female'

if __name__ == "__main__":
    # unittest.main()
    suite1 = unittest.TestLoader().loadTestsFromTestCase(TestCountQF)
    suite2 = unittest.TestLoader().loadTestsFromTestCase(TestAreaQF)
    suite3 = unittest.TestLoader().loadTestsFromTestCase(TestFITarget)
    # suite3 = unittest.TestLoader().loadTestsFromTestCase(TestAlgorithms3)
    complete_suite = unittest.TestSuite([suite1, suite2, suite3])
    unittest.TextTestRunner(verbosity=2).run(complete_suite)



# flake8: noqa
from pysubgroup.algorithms import *
from pysubgroup.binary_target import *
from pysubgroup.constraints import *
from pysubgroup.fi_target import *
from pysubgroup.gp_growth import GpGrowth
from pysubgroup.measures import *
from pysubgroup.model_target import EMM_Likelihood, PolyRegression_ModelClass
from pysubgroup.numeric_target import *
from pysubgroup.refinement_operator import *
from pysubgroup.representations import *
from pysubgroup.subgroup_description import *
from pysubgroup.utils import *
from pysubgroup.visualization import *
